#include "SWFFile.h"
#include "swft.h"
#include "xslt.h"
#include <stdlib.h>
#include <string.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>

using namespace SWF;

bool quiet = false;
bool verbose = false;
bool dump = false;

void usage() {
	fprintf( stderr,
		"\n"PACKAGE_NAME" "PACKAGE_VERSION"\n"
		"     XML-based SWF (Shockwave Flash) processing tool\n"
		"\n"
		"usage: swfmill [<options>] <command>\n"
		"\n"
		"<command> is one of:\n"
		"    swf2xml <in> [<out>]\n"
		"           convert from SWF to XML.\n"
		"           <in> is a single SWF file, or 'stdin'\n"
		"           <out> is a single XML file, or (by default) 'stdout'\n"
		"\n"
		"    xml2swf <in> [<out>]\n"
		"           convert from XML to SWF.\n"
		"           <in> is a single XML file, or 'stdin'\n"
		"           <out> is a single SWF file, or (by default) 'stdout'\n"
		"\n"
		"    xslt <xsl> <in> [<out>"
		//" [<param>*]"
		"]\n"
		"           transform <in> to <out> as described by <xsl>.\n"
		"           <xsl> is the XSLT stylesheet,\n"
		"                 and can use the swft: extension.\n"
		"           <in>  must be some XML (depends on <xsl>)\n"
		"           <out> is either SWF (when it ends in .swf)\n"
		"                 or XML, by default on 'stdout'\n"
//		"           <param>* is a whitespace-separated list of name=value\n"
//		"                 assignments for xsl parameters\n"
		"\n"
		"    library <hdr> <in>* <out>\n"
		"           produce an SWF that contains all <in> files as library items.\n"
		"           <hdr> is '<width>:<height>:<framerate>', eg. 320:240:12\n"
		"           <in>* is a whitespace-separated list of content files\n"
		"                 currently only .jpg and .swf is supported\n"
		"           <out> is either SWF (when it ends in .swf)\n"
		"                 or XML, by default on 'stdout'\n"
		"           the imported assets will be available with attachMovie()\n"
		"           by their basename (filename without path and extension)\n"
		"\n"
		"<option>s are:\n"
		"    -h print this help and quit\n"
		"    -q be quiet, only print errors and warnings\n"
		"    -v verbose output (for debugging)\n"
		"    -d dump SWF data when loaded (for debugging)\n"
		"\n"
		"E-mail bug reports to "PACKAGE_BUGREPORT"\n\n"
	);
}


int swfmill_swf2xml( int argc, char *argv[] ) {
	const char *infile = NULL, *outfile = "stdout";
	bool success = false;
	FILE *in_fp=0, *out_fp=0;
	bool std_in, std_out;

	File input;
	int size, xmlsize;
	char sig;
	Context ctx;

	// parse filenames
	if( argc < 1 || argc > 2 ) {
		usage();
		goto fail;
	}
	infile = argv[0];
	if( argc>1 ) outfile = argv[1];
	
// open files
	std_in = !strncmp( infile, "stdin", 5 );
	std_out = !strncmp( outfile, "stdout", 6 );
	in_fp = std_in ? stdin : fopen( infile, "rb" );
	if( !in_fp ) {
		fprintf(stderr,"ERROR: could not open file %s for reading\n", infile?infile:"stdin" );
		goto fail;
	}
	
	if( !quiet ) fprintf(stderr,"Reading from %s\n", infile );
	
// setup context
	ctx.debugTrace = verbose;

		// treat input as SWF, produce XML
		if( (size = input.load( in_fp, &ctx )) != 0 ) {
			if( dump ) input.dump();
			out_fp = std_out ? stdout : fopen( outfile, "wb" );
			if( !out_fp ) {
				fprintf(stderr,"ERROR: could not open file %s for writing\n", outfile );
				goto fail;
			}
			if( !quiet ) fprintf(stderr,"Writing XML to %s\n", outfile );
			if( (xmlsize = input.saveXML( out_fp, &ctx )) != 0 ) {
				if( !quiet ) fprintf(stderr,"XML saved to %s (%i bytes).\n", outfile, xmlsize );
				success = true;
			}
		}
fail:
	if( in_fp && !std_in ) fclose(in_fp);
	if( out_fp && !std_out ) fclose(out_fp);
	
	return success ? 0 : -1;
}


int swfmill_xml2swf( int argc, char *argv[] ) {
	const char *infile = NULL, *outfile = "stdout";
	bool success = false;
	FILE *in_fp=0, *out_fp=0;
	bool std_in, std_out;

	File input;
	int size, xmlsize;
	char sig;
	Context ctx;

	// parse filenames
	if( argc < 1 || argc > 2 ) {
		usage();
		goto fail;
	}
	infile = argv[0];
	if( argc>1 ) outfile = argv[1];
	
// open files
	std_in = !strncmp( infile, "stdin", 5 );
	std_out = !strncmp( outfile, "stdout", 6 );
	in_fp = std_in ? stdin : fopen( infile, "rb" );
	if( !in_fp ) {
		fprintf(stderr,"ERROR: could not open file %s for reading\n", infile?infile:"stdin" );
		goto fail;
	}
	
	if( !quiet ) fprintf(stderr,"Reading from %s\n", infile );
	
// setup context
	ctx.debugTrace = verbose;
	
		// treat input as XML, produce SWF
	{
		const char *filename = std_in ? "-" : infile ;
		if( input.loadXML( filename, &ctx ) != 0 ) {
			if( dump ) input.dump();
			out_fp = std_out ? stdout : fopen( outfile, "wb" );
			if( !out_fp ) {
				fprintf(stderr,"ERROR: could not open file '%s' for writing\n", outfile );
				goto fail;
			}
			if( !quiet ) fprintf(stderr,"Writing SWF to %s\n", outfile );
			if( (size = input.save( out_fp, &ctx )) != 0 ) {
				success = true;
				if( !quiet ) fprintf(stderr,"SWF saved to %s (%i bytes).\n", outfile, size );
			}
		}
	}
	
fail:
	if( in_fp && !std_in ) fclose(in_fp);
	if( out_fp && !std_out ) fclose(out_fp);
	
	return success ? 0 : -1;
}


int swfmill_xslt( int argc, char *argv[] ) {
	const char *xslfile, *infile, *outfile = "stdout";
	xsltStylesheetPtr transform = NULL;
	xmlDocPtr doc = NULL, doc2 = NULL;
	FILE *out_fp;
	const char *ext;
	Context ctx;
	int size;
	
	// parse filenames
	if( argc < 2 || argc > 3 ) {
		usage();
		goto fail;
	}
	xslfile = argv[0];
	infile = argv[1];
	if( argc>2 ) outfile = argv[2];

	doc = xmlParseFile( infile );
	if( !doc ) {
		fprintf( stderr, "ERROR: input document %s could not be read.\n", infile );
		goto fail;
	}

	swft_register();
	
	transform = xsltParseStylesheetFile( (const xmlChar *)xslfile );
	if( !transform ) {
		fprintf( stderr, "ERROR: stylesheet %s could not be read.\n", xslfile );
		goto fail;
	}
	
	if( !quiet ) fprintf( stderr, "Applying XSLT %s to %s...\n", xslfile, infile );
	doc2 = xsltApplyStylesheet( transform, doc, NULL );
	
	if( !doc2 ) {
		fprintf( stderr, "ERROR: transformation failed.\n" );
		goto fail;
	}	
	
	if( (ext = strrchr(outfile,'.'))!=NULL && !strncmp( ".swf", ext, 4 ) ) {
		if( !quiet ) fprintf( stderr, "Saving to SWF %s\n", outfile);		
		
		File swf;
		swf.setXML( doc2->xmlRootNode, NULL );

		out_fp = fopen( outfile, "wb" );
		if( !out_fp ) {
			fprintf(stderr,"ERROR: could not open file '%s' for writing\n", outfile );
			goto fail;
		}
		if( !quiet ) fprintf(stderr,"Writing SWF to %s\n", outfile );
		if( (size = swf.save( out_fp, &ctx )) != 0 ) {
			if( !quiet ) fprintf(stderr,"SWF saved to %s (%i bytes).\n", outfile, size );
			if( doc ) xmlFreeDoc(doc);
			if( doc2 ) xmlFreeDoc(doc2);
			if( transform ) xsltFreeStylesheet( transform );
			return 0;
		}
	} else {
		bool std_out = !strncmp( outfile, "stdout", 6 );
		const char *file = std_out ? "-" : outfile;
		return xsltSaveResultToFilename( file, doc2, transform, 0 );
	}
	
fail:
	if( doc ) xmlFreeDoc(doc);
	if( doc2 ) xmlFreeDoc(doc2);
	if( transform ) xsltFreeStylesheet( transform );
	return -1;
}

// helper func
xsltStylesheetPtr xsltParseStylesheetMemory( const char *buffer, int size ) {
	xmlDocPtr doc = xmlParseMemory( buffer, size );
	if( !doc ) return NULL;
	xsltStylesheetPtr ret = xsltParseStylesheetDoc( doc );
//	xmlFreeDoc( doc );
	return ret;
}

int swfmill_pack( int argc, char *argv[], const char *xsl ) {
	const char *outfile = "stdout";
	xsltStylesheetPtr transform = NULL;
	xmlDocPtr doc = NULL, doc2 = NULL;
	FILE *out_fp;
	const char *ext;
	Context ctx;
	int size;
	const char *header;
	int width=0, height=0;
	float framerate=0;
	
	// parse filenames
	if( argc < 3 ) {
		usage();
		goto fail;
	}
	outfile = argv[argc-1];

	header = argv[0];
	sscanf( header, "%i:%i:%f", &width, &height, &framerate );
	
	if( width == 0 && height == 0 && framerate == 0 ) {
		fprintf(stderr,"WARNING: first argument to 'library' must be header (eg. 320:240:12).\nignoring '%s'\n", header );
		width = 320; height=240; framerate=12;
	}
		
	doc = xslt_pack( argc-2, &argv[1], width*20, height*20, framerate );
	if( !doc ) {
		fprintf( stderr, "no files to import.\n" );
		goto fail;
	}

	swft_register();
	
	// main transform
	transform = xsltParseStylesheetMemory( xsl, strlen(xsl) );
	if( !transform ) {
		fprintf( stderr, "FATAL: internal stylesheet could not be read.\n" );
		goto fail;
	}
	doc2 = xsltApplyStylesheet( transform, doc, NULL );
	if( !doc2 ) {
		fprintf( stderr, "ERROR: transformation failed.\n" );
		goto fail;
	}	

	// setup context
		ctx.debugTrace = verbose;
	
	// output
	if( (ext = strrchr(outfile,'.'))!=NULL && !strncmp( ".swf", ext, 4 ) ) {
		if( !quiet ) fprintf( stderr, "Saving to SWF %s\n", outfile);		
		
		File swf;
		swf.setXML( doc2->xmlRootNode, &ctx );

		out_fp = fopen( outfile, "wb" );
		if( !out_fp ) {
			fprintf(stderr,"ERROR: could not open file '%s' for writing\n", outfile );
			goto fail;
		}
		if( !quiet ) fprintf(stderr,"Writing SWF to %s\n", outfile );
		if( (size = swf.save( out_fp, &ctx )) != 0 ) {
			if( !quiet ) fprintf(stderr,"SWF saved to %s (%i bytes).\n", outfile, size );
			if( doc ) xmlFreeDoc(doc);
			if( doc2 ) xmlFreeDoc(doc2);
			if( transform ) xsltFreeStylesheet( transform );
			return 0;
		}
	} else {
		bool std_out = !strncmp( outfile, "stdout", 6 );
		const char *file = std_out ? "-" : outfile;
		return xsltSaveResultToFilename( file, doc2, transform, 0 );
	}
	
fail:
	if( doc ) xmlFreeDoc(doc);
	if( doc2 ) xmlFreeDoc(doc2);
	if( transform ) xsltFreeStylesheet( transform );
	return -1;
}

int main( int argc, char *argv[] ) {
	char *command = NULL;
	
// parse args
	int i=1;
	for( ; i<argc && command==NULL; i++ ) {
		if( argv[i][0] == '-' ) {
			for( int j=1; j<strlen(argv[i]); j++ ) {
				switch( argv[i][j] ) {
					case 'q':
						quiet = true;
						break;
					case 'v':
						verbose = true;
						break;
					case 'd':
						dump = true;
						break;
					case 'h':
						usage();
						goto fail;
						break;
					default:
						fprintf(stderr,"ERROR: unknown option %c\n",argv[i][j]);
						usage();
						goto fail;
				}
			}
		} else {
			if( !command ) command=argv[i];
		}
	}
	
	if( !command ) {
		usage();
		goto fail;
	}
	
	if( !quiet ) fprintf( stderr, "%s %s\n", PACKAGE_NAME, PACKAGE_VERSION );

	if( !strcmp( command, "swf2xml" ) ) {
		return swfmill_swf2xml( argc-i, &argv[i] );
	} else if( !strcmp( command, "xml2swf" ) ) {
		return swfmill_xml2swf( argc-i, &argv[i] );
	} else if( !strcmp( command, "xslt" ) ) {
		return swfmill_xslt( argc-i, &argv[i] );
	} else if( !strcmp( command, "library" ) ) {
		return swfmill_pack( argc-i, &argv[i], xslt_library );
	} else {
		fprintf(stderr,"ERROR: unknown command %s\n", command );
		usage();
		goto fail;
	}
		
fail:
	return -1;
}

